// Copyright (c) 2001, 2002 Zope Foundation and Contributors.
// All Rights Reserved.
//
// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE

package main
// Bits that should be in ZODB		XXX -> zodb

import (
	"context"
	"fmt"
	"sync"

	"lab.nexedi.com/kirr/neo/go/zodb"

	pickle "github.com/kisielk/og-rek"
)

// Object is the interface that every in-RAM object representing any ZODB object implements.
type Object interface {
	// XXX make methods private, e.g. _pJar ?
	PJar()    *Connection
	POid()    zodb.Oid
	PSerial() zodb.Tid

	// PActivate brings object to live state.
	//
	// It requests to persistency layer that in-RAM object data to be present.
	// If object state is not in RAM - it is loaded from the database.
	//
	// On successful return the object data is either the same as in the
	// database or, if this data was previously modified by user of
	// object's jar, that modified data.
	//
	// Object data must be accessed only after corresponding PActivate
	// call, which marks that object's data as being in use.
	PActivate(ctx context.Context) (activated bool, _ error)	// XXX + -> nuse ?

	// PDeactivate indicates that corresponding PActivate caller finished access to object's data.
	//
	// As PActivate makes sure object's data is present in-RAM, PDeactivate
	// tells persistency layer that this data is no longer used by
	// corresponding PActivate caller.
	//
	// Note that it is valid to have several concurrent uses of object
	// data, each protected with corresponding PActivate/PDeactivate pair:
	// As long as there is still any PActivate not yet compensated with
	// corresponding PDeactivate, object data will assuredly stay alive in RAM.
	//
	// Besides exotic cases, the caller thus must not use object's data
	// after PDeactivate call.
	PDeactivate()	// XXX + -> nuse ?

	// PInvalidate requests in-RAM object data to be discarded.
	//
	// Irregardless of whether in-RAM object data is the same as in the
	// database, or it was modified, that in-RAM data must be forgotten.
	//
	// PInvalidate must not be called while there is any in-progress
	// object's data use (PActivate / PDeactivate).
	PInvalidate()
}

// PyObject is the interface that every in-RAM object representing Python ZODB object implements.
type PyObject interface {
	Object

	PyClass() pickle.Class // python class of this object
	PyState() interface{}  // object state. python passes this to pyclass.__new__().__setstate__()

}

// object is common base for in-RAM representation of ZODB object.
type object struct {
	jar	*Connection
	oid	zodb.Oid
	serial	zodb.Tid

	refcnt	int32
}

func (obj *object) PJar() *Connection	{ return obj.jar	}
func (obj *object) POid() zodb.Oid	{ return obj.oid	}
func (obj *object) PSerial() zodb.Tid	{ return obj.serial	}

// pyObject is common base for in-RAM representation of ZODB Python objects.
type pyObject struct {
	object

	pyclass pickle.Class
	pystate interface{}
	loaderr error		// if there was error at state loading
	ready   chan struct{}	// activation complete
}

func (pyobj *pyObject) PyClass() pickle.Class	{ return pyobj.pyclass	}
func (pyobj *pyObject) PyState() interface{}	{ return pyobj.pystate	}


// Connection represents a view of ZODB database.
//
// The view is representing state of ZODB objects as of `at` transaction.
//
// XXX Connection changes are private and are isolated from changes in other Connections.
//
// XXX Connection, and {Py}Object methods that relate to it, are not safe for
// modifications from multiple goroutines simultaneously.
//
// XXX ^^^ better must be safe - use case: e.g. prefetch.
type Connection struct {
	stor	zodb.IStorage	// underlying storage
	at	zodb.Tid	// current view of database

	// {} oid -> obj
	//
	// rationale:
	//
	// on invalidations: we need to go oid -> obj and invalidate it.
	// -> Connection need to keep {} oid -> obj.
	// -> we can use that {} when loading a persistent Ref twice to get to the same object.
	//
	// however: if Connection keeps strong link to obj, just
	// obj.PDeactivate will not fully release obj if there are no
	// references to it from other objects:
	//
	//	- deactivate will release obj state (ok)
	//	- but there will be still reference from connection `oid -> obj` map to this object.
	//
	// -> we can solve it by using "weak" pointers in the map.
	//
	// NOTE we cannot use regular map and arbitrarily manually "gc" entries
	// there periodically: since for an obj we don't know whether other
	// objects are referencing it, we can't just remove obj's oid from
	// the map - if we do so and there are other live objects that
	// reference obj, user code can still reach obj via those
	// references. On the other hand, if another, not yet loaded, object
	// also references obj and gets loaded, traversing reference from
	// that loaded object will load second copy of obj, thus breaking 1
	// object in db <-> 1 live object invariant:
	//
	//	A  →  B  →  C
	//	↓           |
	//      D <--------- - - -> D2 (wrong)
	//
	// - A activate
	// - D activate
	// - B activate
	// - D gc, A still keeps link on D
	// - C activate -> it needs to get to D, but D was removed from objtab
	//   -> new D2 is wrongly created
	//
	// that's why we have to depend on Go's GC to know whether there are
	// still live references left or not. And that in turn means finalizers
	// and thus weak references.
	//
	// some link on the subject:
	// https://groups.google.com/forum/#!topic/golang-nuts/PYWxjT2v6ps
	//
	// NOTE2 finalizers don't run on when they are attached to an object in cycle.
	// Hopefully we don't have cycles with ZBtree/ZBucket	XXX verify this
	objmu  sync.Mutex
	objtab map[zodb.Oid]*WeakRef	// oid -> WeakRef(PyObject)
	//objtab map[zodb.Oid]interface{}	// oid -> WeakRef(pyObject) | loadInProgress
}

/*
// loadInProgress entry in Conn.objtab tells users, that try to get the entry,
// that another goroutine is already in progress of loading it.
type loadInProgress struct {
	ready chan struct{} // closed when loading finishes

	// result of the load
	pyobj interface{} // XXX -> pyObject iface
	err   error
}
*/

// Get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created and filled with object's class loaded from the database.
//
// The scope of the object returned is the Connection.	XXX ok?
//
// The object's data is not neccessarily loaded after Get returns. Use
// PActivate to make sure the object ifs fully loaded.
func (conn *Connection) Get(ctx context.Context, oid zodb.Oid) (interface{} /*PyObject*/, error) {
	conn.objmu.Lock()
	wobj := conn.objtab[oid]
	var xobj interface{}
	if wobj != nil {
		xobj = wobj.Get()
	}
	conn.objmu.Unlock()

	// object was already there in objtab.
	if xobj != nil {
		return xobj, nil
	}

	// object is not there in objtab - raw load it, get its class -> get(pyclass, oid)
	pyclass, pystate, serial, err := conn.loadpy(ctx, oid)
	if err != nil {
		return nil, err		// XXX errctx
	}

	xobj, err = conn.get(pyclass, oid)
	if err != nil {
		return nil, err
	}

	// XXX we are dropping just loaded pystate. Usually Get should be used
	// to only load root object, so maybe that is ok.
	//
	// TODO -> use (pystate, serial) to activate.
	_, _ = pystate, serial
	return xobj, nil
}

/*
	conn.objmu.Lock()		// XXX -> rlock
	objentry := conn.objtab[oid]

	// someone else is already loading the object.
	// we have to wait for that someone to finish the load.
	if load, ok := objentry.(*loadInProgress); ok {
		conn.objmu.Unlock()
		select {
		case <-ctx.Done():
			return nil, ctx.Err()	// XXX err ctx
		case <-load.ready:
			// XXX check pyclass match
			return load.pyobj, load.err
		}
	}

	// noone else is loading the object. Investigate whether it is simply not
	// in objtab, or it is lost weakref, and if so we have to initiate the load.
	var xobj interface{}
	if objentry != nil {
		wobj := objentry.(*WeakRef)
		xobj = wobj.Get()
	}
	if xpyobj != nil {
		// the object is already there
		conn.objmu.Unlock()
		return xpyobj, nil	// XXX + activate?
	}

	// the object is not there, we have to (re)load it.
	load := &loadInProgress{ready: make(chan struct)}
	conn.objTab[oid] = load
	conn.objmu.Unlock()

	pyobj, err = conn.loadpy(ctx, oid)
	load.pyobj = pyobj
	load.err   = err
	close(load.ready)

	// replace objtab entry with WeakRef
	conn.objmu.Lock()
	defer conn.objmu.Unlock()
	if x := conn.objtab[oid]; x != load {
		panic(fmt.Sprintf("zodb.Conn: objtab[%s]: load finished, but it is: %#v", oid, x)
	}
	if err != nil {
		// it was loading error - remove the entry not to cause uncontrolled objtab growth.
		// if this object is accessed next time - the load will be retried.
		delete(conn.objtab[oid])
	} else {
		conn.objtab[oid] = NewWeakRef(pyobj)
	}
*/

// wrongClassError is the error cause returned when object's class is not what was expected.
type wrongClassError struct {
	want, have pickle.Class
}

func (e *wrongClassError) Error() string {
	return fmt.Sprintf("wrong class: want %q; have %q", e.want, e.have)
}

// get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created according to specified class.
//
// The object's data is not neccessarily loaded after get returns. Use
// PActivate to make sure the object is fully loaded.
//
// XXX object scope.
//
// use-case: in ZODB references are (pyclass, oid), so new ghost is created
// without further loading anything.
func (conn *Connection) get(pyclass pickle.Class, oid zodb.Oid) (PyObject, error) {
	conn.objmu.Lock()		// XXX -> rlock
	wobj := conn.objtab[oid]
	var pyobj PyObject
	checkClass := false
	if wobj != nil {
		if xobj := wobj.Get(); xobj != nil {
			pyobj = xobj.(PyObject)
		}
	}
	if pyobj == nil {
		pyobj = conn.newGhost(pyclass, oid)
		conn.objtab[oid] = NewWeakRef(pyobj)
	} else {
		checkClass = true
	}
	conn.objmu.Unlock()

	if checkClass {
		if cls := pyobj.PyClass(); pyclass != cls {
			return nil, &zodb.OpError{
				URL:  conn.stor.URL(),
				Op:   fmt.Sprintf("@%s: get", conn.at), // XXX abuse
				Args: oid,
				Err:  &wrongClassError{pyclass, cls},
			}
		}
	}

	return pyobj, nil
}


/*
{
	objentry := conn.objtab[oid]

	// someone else is already loading the object.
	// we have to wait for that someone to finish the load.
	if load, ok := objentry.(*loadInProgress); ok {
		conn.objmu.Unlock()
		select {
		case <-ctx.Done():
			return nil, ctx.Err()	// XXX err ctx
		case <-load.ready:
			// XXX check pyclass match
			return load.pyobj, err
		}
	}

	// noone else is loading the object. Investigate whether it is simply not
	// in objtab, or it is lost weakref, and if so we have to initiate the load.
	var xobj interface{}
	if objentry != nil {
		wobj := objentry.(*WeakRef)
		xobj = wobj.Get()
	}
	if xpyobj != nil {
		// the object is already there
		conn.objmu.Unlock()
		return xpyobj, nil	// XXX + activate?
	}

	// the object is not there, we have to (re)load it.
	load := &loadInProgress{ready: make(chan struct)}
	conn.objTab[oid] = load
	conn.objmu.Unlock()

	pyobj, err = conn.loadpy(ctx, oid)
	load.pyobj = pyobj
	load.err   = err
	close(load.ready)

	// replace objtab entry with WeakRef
	conn.objmu.Lock()
	defer conn.objmu.Unlock()
	if x := conn.objtab[oid]; x != load {
		panic(fmt.Sprintf("zodb.Conn: objtab[%s]: load finished, but it is: %#v", oid, x)
	}
	if err != nil {
		// it was loading error - remove the entry not to cause uncontrolled objtab growth.
		// if this object is accessed next time - the load will be retried.
		delete(conn.objtab[oid])
	} else {
		conn.objtab[oid] = NewWeakRef(pyobj)
	}
}
*/

/*
	// XXX -> loadpy
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return nil, err
	}

	pyclass, pystate, err := zodb.PyData(buf.Data).Decode()
	if err != nil {
		return nil, err	// XXX err ctx
	}

	buf.Release()

	return &pyObject{
		object:  object{jar: conn,oid: oid, serial: serial},
		pyclass: pyclass,
		pystate: pystate,
	}, nil
}
*/

/*
func (conn *Connection) load(ctx context.Context, oid zodb.Oid) (*pyObject, error) {
	// XXX
}
*/

func (conn *Connection) loadpy(ctx context.Context, oid zodb.Oid) (pyclass pickle.Class, pystate interface{}, serial zodb.Tid, _ error) {
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return pickle.Class{}, nil, 0, err
	}

	pyclass, pystate, err = zodb.PyData(buf.Data).Decode()
	if err != nil {
		return pickle.Class{}, nil, 0, err	// XXX err ctx
	}

	buf.Release()

	return pyclass, pystate, serial, nil
}

// path(class) -> new(pyobj)
var classTab = make(map[string]func(*pyObject)PyObject)

// registerClass registers python class to be transformed to Go instance
// created via classNew.
//
// must be called from global init().
func registerClass(classPath string, classNew func(*pyObject)PyObject) {
	classTab[classPath] = classNew
}

// newGhost creates new ghost object corresponding to pyclass and oid.
func (conn *Connection) newGhost(pyclass pickle.Class, oid zodb.Oid) PyObject {
	pyobj := &pyObject{
		object:  object{jar: conn, oid: oid, serial: 0},
		pyclass: pyclass,
		pystate: nil,
	}

	// switch on pyclass and transform e.g. "zodb.BTree.Bucket" -> *ZBucket
	classNew := classTab[pyclass.Module + "." + pyclass.Name]
	if classNew == nil {
		return pyobj	// XXX or return error here?
	}

	return classNew(pyobj)
}



// ----------------------------------------

/*
func (obj *object) PActivate(ctx context.Context) error {
	// XXX
}

func (obj *object) PDeactivate() {
	// XXX
}

// PInvalidate
*/

// object's pactivate & friends that only manage base activation state, without actually loading data.

func (obj *object) pactivate(ctx context.Context) error {
	nuse := atomic.AddInt32(&obj.refcnt, +1)
	if nuse == 1 {
		// we become responsible for loading object's data.
		// XXX check .tryToKeepInRam
	}
}

func (obj *object) pdeactivate() {
	nuse := atomic.AddInt32(&obj.refcnt, -1)
	if nuse < 0 {
		panic("pdeactivate: nuse < 0")
	}
	if nuse > 0 {
		return // users still left
	}

	// no users left. Let's see whether we should transition this object to ghost.
	// TODO state=modified -> don't drop.

	if drop {
		obj.serial  = 0
	}
}


// PActivate implements Object.
func (pyobj *pyObject) PActivate(ctx context.Context) (bool, error) {
	load := pyobj.object.pactivate()
	if !load {
		// someone else is already activated/activating the object.
		// wait for its loading to complete and we are done.
		select {
		case <-ctx.Done:
			return false, ctx.Err()		// XXX err ctx
		case <-pyobj.ready:
			return (pyobj.loaderr == nil), pyobj.loaderr		// XXX err ctx?
		}
	}

	// we become responsible for loading the object
	pyclass, pystate, serial, err := pyobj.jar.loadpy(ctx, pyobj.oid)
	if err == nil && pyclass != pyobj.pyclass {
		// complain pyclass changed
		// (both ref and object data uses pyclass so it indeed can be different)
		err = &wrongClassError{want: pyobj.pyclass, have: pyclass} // XXX + err ctx
		pystate = nil
	}

	pyobj.serial  = serial
	pyobj.pystate = pystate
	pyobj.loaderr = err
	close(pyobj.ready)
	return err	// XXX err ctx
}


// PDeactivate implements Object.
func (pyobj *pyObject) PDeactivate() {
	drop := pyobj.object.pdeactivate()
	if !drop {
		return
	}

	// we have to drop pyobject state
	// XXX locking?
	pyobj.pystate = nil
	pyobj.loaderr = nil
	pyobj.ready   = make(chan struct{})

	return drop
}

// XXX pyobj.PInvalidate()	= deactivate without checking if state != modified


